% markerAccuracySimulator (configFilePath)
%
% Inputs:
% - configFilePath: path to the conficuration file 
%   (optionnal, default: 'config.ini')
%
% This script simulates the measurement noise of:
% - a marker (rigid body) and
% - points physically situated on it.
% 
% Example of such points can be a pivot point, a nick used for tool zeroing, etc.
% The simulation returns the expected measurement noise (std, rms) at these points.
%
% It is optionally possible to define a reference marker in order to 
% simulate the error of the points expressed in the reference marker referential.
% Please use 'configWithRef.ini' configuration file to test this feature.
%
% Configuration of the simulation is performed via the configFilePath file.
%
% Note that the referential of the device is defined at the mid-point
% between the two cameras (and not the left camera as in the real device)
%
%
% This script has been successfully run with Matlab 2015b (maci64)
%
% Author: Gaetan Marti
% September 22nd, 2017; Last revision: September 27th, 2017
% Version: 1.2
%
% Copyright (c) 2017, Atracsys LLC
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are not permitted
%
% Code may be adapted for your own needs but cannot be redistributed
%
%
% Releases history:
% - September 26th: Adding marker tol, output display and corrections of some typos.
% - September 27th: Adding the reference marker feature
%

function markerAccuracySimulator(configFilePath)

    % Check the presence of a specific section/entry in the ini file
    function checkSectionEntry (section, entry)
        if ~isfield (config, section)
            error ('No [%s] section in ''config.ini'' configuration file', section);
        end
        if ~isfield (config.(section), entry)
            error ('No %s=... entry in [%s] section',entry, section);
        end
    end

    % Get a section/entry in the ini file and return default if not present
    % toNum forces the entry to be double.
    function ret = getEntry (section, entry, default, toNum)
        ret = default;
        if ~isfield (config, section) || ~isfield (config.(section), entry)
            return
        end
        if exist ('toNum', 'var') && toNum
            ret = str2double (config.(section).(entry));
        else
            ret = config.(section).(entry);
        end
    end

    % Adding tolerance is complex as it depends on the manufacturing
    % tools. For simplification purposes, it is defined as a uniform noise.
    function ret = addTolError (points, tol)
        ret = points;
        if norm (tol) == 0.0
            return;
        end
        ret = points + 2*tol .* rand (size(points)) - tol;
    end

    if ~exist ('configFilePath', 'var')
        configFilePath = 'config.ini';
    end

    % Load the ini file
    if ~exist (configFilePath, 'file')
        error ('Missing ''%s'' configuration file', configFilePath);
    end
    config = ini2struct (configFilePath);

    % Get the number of samples
    samples = getEntry ('parameters','samples', 1000, true);

    % Get the device type
    checkSectionEntry ('parameters','device');
    c = const();
    if strcmpi(config.parameters.device, 'FTK250')
        tracker = trackerGeometry( c.FUSIONTRACK_250 );
    elseif strcmpi(config.parameters.device, 'FTK500')
        tracker = trackerGeometry( c.FUSIONTRACK_250 );
    elseif strcmpi(config.parameters.device, 'STK180')
        tracker = trackerGeometry( c.FUSIONTRACK_250 );
    else
        error ('Invalid device type');
    end

    % --------- measured marker --------- 
    
    % Load the geometry file of the marker
    if isfield (config, 'geometry') && isfield (config.geometry, 'file')
        warning ('''file'' entry should now be placed in ''marker'' section.');
        [~, markerFiducialPos] = loadMarkerIniFile(config.geometry.file);
    else
        checkSectionEntry ('marker','file');
        [~, markerFiducialPos] = loadMarkerIniFile(config.marker.file);
    end
    
    % Get the rotation of the marker
    markerRotationVec = [getEntry('marker','rotx',0.0, true); ...
                         getEntry('marker','roty',0.0, true); ...
                         getEntry('marker','rotz',0.0, true)];
    markerRotation = rodrigues (markerRotationVec);

    % Get the tolerance of the marker
    markerTol = [getEntry('marker','tolx',0.0, true); ...
                 getEntry('marker','toly',0.0, true); ...
                 getEntry('marker','tolz',0.0, true)];

    % Get the position of the marker
    checkSectionEntry ('marker','posx');
    checkSectionEntry ('marker','posy');
    checkSectionEntry ('marker','posz');
    markerPosition = [str2double(config.marker.posx); ...
                      str2double(config.marker.posy); ...
                      str2double(config.marker.posz)];

    % --------- reference marker --------- 
    
    if isfield (config, 'markerref')
    
        checkSectionEntry ('markerref','file');
        [~, refMarkerFiducialPos] = loadMarkerIniFile(config.markerref.file);
        
        % Get the rotation of the marker
        refMarkerRotationVec = [getEntry('markerref','rotx',0.0, true); ...
                                getEntry('markerref','roty',0.0, true); ...
                                getEntry('markerref','rotz',0.0, true)];
        refMarkerRotation = rodrigues (refMarkerRotationVec);

        % Get the tolerance of the marker
        refMarkerTol = [getEntry('markerref','tolx',0.0, true); ...
                        getEntry('markerref','toly',0.0, true); ...
                        getEntry('markerref','tolz',0.0, true)];

        % Get the position of the marker
        checkSectionEntry ('markerref','posx');
        checkSectionEntry ('markerref','posy');
        checkSectionEntry ('markerref','posz');
        refMarkerPosition = [str2double(config.markerref.posx); ...
                            str2double(config.markerref.posy); ...
                            str2double(config.markerref.posz)];  
    end
    
    % --------- points on measured marker --------- 
                  
    % Check that at least a virtual point is present
    checkSectionEntry ('point00','posx');

    calibPoints = [];

    plotType = {'r.', 'b.', 'c.', 'k.', 'rx', 'bx', 'cx', 'kx'};

    for cpNb = 0:1000
        section = sprintf ('point%.2d',cpNb);
        if ~isfield (config, section)
            break;
        end

        % Get the position of the point in the marker referential
        x = getEntry(section,'posx',0.0,true);
        y = getEntry(section,'posy',0.0,true);
        z = getEntry(section,'posz',0.0,true);

        calibPoints = [calibPoints, [x;y;z]]; %#ok<AGROW>

        % Get point name and plot type
        calibPointsInfo{cpNb+1}.name = getEntry (section,'name', section); %#ok<AGROW>
        calibPointsInfo{cpNb+1}.plot = plotType {mod(cpNb,size(plotType,2))+1};         %#ok<AGROW>
    end

    % --------- start of simulation --------- 
    
    % Ground truth: absolute position of measured points
    calibPointInTrackerReference = markerRotation * calibPoints + repmat (markerPosition,1,cpNb);

    % Simulate tracking error
    markerPose      = markerFiducialPos;
    calibPointError = zeros (cpNb, 3,samples);
    markerRegError  = zeros (samples,1);
    
    if isfield (config, 'markerref')
        
        % Express the points in the reference marker
        % Ex: transformation after a registration process
        calibPointInRefMarkerReference = ...
            refMarkerRotation' * calibPointInTrackerReference - ...
            repmat (refMarkerRotation' * refMarkerPosition, 1, cpNb); 
        
        refMarkerPose      = refMarkerFiducialPos;
        refCalibPointError = zeros (cpNb, 3,samples);
        refMarkerRegError  = zeros (samples,1);
    end

    % Perform the computation
    for s = 1:samples
        
        % Simulated marker pose
        for i=1:size( markerFiducialPos, 2 )
            markerPose(:,i) = markerRotation * addTolError(markerFiducialPos(:,i), markerTol) + markerPosition;
        end
        noisyMarkerPose = simulateNoise( markerPose, tracker );
        [ noisyRot, noisyTrans, regMarkerError ] = computeMarker( markerFiducialPos, noisyMarkerPose );
        markerRegError(s) = regMarkerError;
        
        % Simulated reference marker pose
        if isfield (config, 'markerref')
            for i=1:size( refMarkerFiducialPos, 2 )
                refMarkerPose(:,i) = refMarkerRotation * addTolError(refMarkerFiducialPos(:,i), refMarkerTol) + refMarkerPosition;
            end
            noisyRefMarkerPose = simulateNoise( refMarkerPose, tracker );
            [ noisyRefRot, noisyRefTrans, regRefMarkerError ] = computeMarker( refMarkerFiducialPos, noisyRefMarkerPose );
            refMarkerRegError(s) = regRefMarkerError;
        end
        
         % Compute the differrence between the effective point position
        % and the measured one in tracker referencial
        noisyCalibPoints = noisyRot * calibPoints + repmat (noisyTrans,1,cpNb);
        
        if isfield (config, 'markerref')
            
            noisyCalibPointsInRefMarkerReference = ...
                noisyRefRot' * noisyCalibPoints - ...
                repmat (noisyRefRot' * noisyRefTrans, 1, cpNb); 
            
            for p=1:cpNb
                calibPointError(p,:,s) = noisyCalibPointsInRefMarkerReference(:,p) - calibPointInRefMarkerReference(:,p);
            end
            
        else
                
            for p=1:cpNb
                calibPointError(p,:,s) = noisyCalibPoints(:,p) - calibPointInTrackerReference(:,p);
            end

        end
    end

    % Displa the typical registration error for the marker
    figure (1)
    clf;
    hold on;
    
        title ('Marker(s) registration error');
        disp ('Marker(s) registration error:');

        legends{1} = sprintf (' Measured marker error: (mean: %.3f m, std: %.3f mm, rms: %.3f mm)', ...
            mean (markerRegError), std(markerRegError), rms(markerRegError));
        disp(legends{1});
        if isfield (config, 'markerref')
            hist ([markerRegError, refMarkerRegError]);
            
            legends{2} = sprintf (' Reference marker error: (mean: %.3f m, std: %.3f mm, rms: %.3f mm)', ...
            mean (refMarkerRegError), std(refMarkerRegError), rms(refMarkerRegError));
            disp(legends{2});
        else
            hist (markerRegError);
        end
        
        legend (legends);
        clear legends;
        
    hold off;

    % Display the resulting error centered at the virutal points
    figure (2)
    clf;
    hold on;
    axis equal;

    disp ('Noise at points:');

    for p=1:cpNb

        x = squeeze(calibPointError(p,1,:));
        y = squeeze(calibPointError(p,2,:));
        z = squeeze(calibPointError(p,3,:));
        legends{p} = sprintf (' ''%s'', std:(%.3f,%.3f,%.3f), rms:(%.3f,%.3f,%.3f)', ...
                        calibPointsInfo{p}.name, ...
                        std(x), std(y), std(z), rms(x), rms(y), rms(z)); %#ok<AGROW>
        plot3(x,y,z, calibPointsInfo{p}.plot);
        disp (legends{p});
    end
    legend (legends);
    xlabel('X [mm]');
    ylabel('Y [mm]');
    zlabel('Z [mm]');
    title (sprintf ('Simulation of noise at pos: (%.0f mm, %.0f mm, %.0f mm), rot: %.1f deg ', ...
        markerPosition (1), markerPosition (2), markerPosition (3), rad2deg(norm(markerRotationVec))));

    hold off;

end
